#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""inspire.py
Collection of functions useful for the INSPIRE challenge.

"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import logging
import os
import pprint as pp
import collections
import re
import shlex
import urllib2
import progressbar as pb

from . import common
from . import edit_distance

UTF8_NORMALIZATION = 'NFD'

BASE_URL = 'http://143.167.9.43:5000'


def _get_url(url):
    try:
        u = urllib2.urlopen(url)
        resp = u.read()
        u.close()
    except urllib2.HTTPError:
        logging.error('Could not get: \n'
                      '{}\n'
                      'Please check your internet connection.\n'
                      'If problem persists inform the challenge organizers.'.format(url))
        return '{}'

    return resp


def _download_url(url, filename=None):
    try:
        u = urllib2.urlopen(url)
    except urllib2.HTTPError:
        logging.error('Could not download: \n'
                      '{}\n'
                      'Please check your internet connection.\n'
                      'If problem persists inform the challenge organizers.'.format(url))
        return None

    meta = u.info()

    reported_filename = shlex.split(re.findall(r'filename=(\S+)',
                                               meta.getheaders('Content-Disposition')[0])[0])[0]

    filename = filename or reported_filename

    if os.path.isfile(filename):
        u.close()
        return filename

    f = open(filename, 'wb')
    file_size = int(meta.getheaders("Content-Length")[0])

    widgets = ['Downloading: ', pb.Percentage(), ' ', pb.Bar(),
               ' ', pb.ETA(), ' ', pb.FileTransferSpeed()]

    pbar = pb.ProgressBar(widgets=widgets, maxval=file_size).start()

    file_size_dl = 0
    block_sz = 8192
    while True:
        download_buffer = u.read(block_sz)
        if not download_buffer:
            break

        file_size_dl += len(download_buffer)
        f.write(download_buffer)

        pbar.update(file_size_dl)

    pbar.finish()

    print('Finished downloading {}'.format(filename))

    f.close()
    u.close()
    return filename


class MyPrettyPrinter(pp.PrettyPrinter):
    _escape = dict((q, dict(((c, unicode(repr(chr(c)))[1:-1])
                             for c in range(32) + [ord('\\')] +
                             range(128, 161)),
                            **{ord(q): u'\\' + q}))
                   for q in ["'", '"'])

    def format(self, obj, context, maxlevels, level):
        if type(obj) is unicode:
            q = "'" if "'" not in obj or '"' in obj \
                else '"'
            return ("u" + q + obj.translate(self._escape[q]) +
                    q, True, False)
        return pp.PrettyPrinter.format(
            self, obj, context, maxlevels, level)


def pprint(obj, **kwargs):
    printer = MyPrettyPrinter(**kwargs)
    printer.pprint(obj)


class EvaluationSetting(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __repr__(self):
        return json.dumps(self, indent=2)

    def download_dataset(self, filename=None):
        return _download_url('{}/download/dataset/{}'.format(BASE_URL, self['dataset_id']), filename=filename)

    def download_lexicon(self, filename=None):
        return _download_url('{}/download/lexicon/{}'.format(BASE_URL, self['lexicon_id']), filename=filename)


class Dataset(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __repr__(self):
        return json.dumps(self, indent=2)


class Lexicon(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.upper())

    def __setitem__(self, key, value):
        return dict.__setitem__(self, key.upper(), value)

    def __contains__(self, key):
        return dict.__contains__(self, key.upper())

    def __repr__(self):
        return json.dumps(self, indent=2)


class Submission(dict):
    def __init__(self, email='', evaluation_setting=None, description=''):
        evaluation_setting_id = evaluation_setting['id'] if evaluation_setting is not None else None

        self.update({'metadata': {'email': email,
                                  'description': description,
                                  'evaluation_setting_id': evaluation_setting_id},
                     'tokens': {}})

    def where_task(self, token_id, confusion_probability):
        self['tokens'].setdefault(token_id, {})['where'] = list(confusion_probability)

    def what_task(self, token_id, index, phonemes, phonemes_probability):
        self['tokens'].setdefault(token_id, {}) \
            .setdefault('what', {}) \
            .setdefault(str(index), {})[phonemes] = phonemes_probability

    def full_task(self, token_id, pronunciation, pronunciation_probability):
        """Provide the prediction of the full task.

        This is function is used to predict the porbability of a given pronunciation being reported for a given token.

        :param token_id: The token for which the prediction is being provided
        :param pronunciation: The pronunciation for which the prediction is being provided
        :param pronunciation_probability: The probability of the pronunciation for the given token
        """

        if not 0. < pronunciation_probability < 1.:
            logging.warning()

        self['tokens'].setdefault(token_id, {}) \
            .setdefault('full', {})[pronunciation] = pronunciation_probability

    def save(self, filename):
        """Save the submission into a file.

        :param filename: where to save the submission in JSON format
        """
        with open(filename, 'w') as f:
            json.dump(self, f, indent=2)

    def submit(self, password=''):
        """Submits the participation to the web site.

        The passwords is sent as plain text.

        :return: the evaluation results.
        """

        data = json.dumps({'email': self['metadata']['email'],
                           'password': password,
                           'submission': self})
        req = urllib2.Request('{}/submit_with_login'.format(BASE_URL),
                              data, {'Content-Type': 'application/json'})
        f = urllib2.urlopen(req)
        resp = f.read()
        f.close()

        try:
            response = json.loads(resp)

        except urllib2.HTTPError as e:
            logging.error('Error while submitting the participation. {}'.format(e))
            return {}

        if 'error' in response:
            logging.error('Error while processing the participation. {}'.format(response['error']))
            return {}

        return response

    def evaluate(self, password=''):
        """Submits the participation to the web site.

        The passwords is sent as plain text.

        :return: the evaluation results.
        """

        data = json.dumps({'email': self['metadata']['email'],
                           'password': password,
                           'submission': self})
        req = urllib2.Request('{}/evaluate_with_login'.format(BASE_URL),
                              data, {'Content-Type': 'application/json'})
        f = urllib2.urlopen(req)
        resp = f.read()
        f.close()

        try:
            response = json.loads(resp)

        except urllib2.HTTPError as e:
            logging.error('Error while submitting the participation. {}'.format(e))
            return {}

        if 'error' in response:
            logging.error('Error while processing the participation. {}'.format(response['error']))
            return {}

        return response

    def __repr__(self):
        return json.dumps(self, indent=2)


def get_evaluation_setting(setting_id=None):
    evaluation_setting_url = '{}/evaluation_setting'.format(BASE_URL)
    if setting_id is not None:
        evaluation_setting_url += '/{}'.format(setting_id)

    return EvaluationSetting(json.loads(_get_url(evaluation_setting_url)))


def load_wordlist(wordlist_filename):
    return set(common.parse_wordlist(wordlist_filename))


def load_lexicon(lexicon_filename):
    return Lexicon({k: [pron.split(' ') for pron in v]
                    for k, v in common.parse_dictionary(lexicon_filename).items()})


def loads_lexicon(lexicon_file):
    return Lexicon({k: [pron.split(' ') for pron in v]
                    for k, v in common.parses_dictionary(lexicon_file).items()})


def load_dataset(dataset_filename):
    with open(dataset_filename) as dataset_file:
        return loads_dataset(dataset_file)

    return {}


def loads_dataset(dataset_file):
    return Dataset(json.load(dataset_file))


def _combine_dicts(*args):
    all_dict = []
    for arg in args:
        all_dict += list(arg.items())

    return dict(all_dict)


def get_edit_scripts(pron_a, pron_b, edit_costs=(1.0, 1.0, 1.0)):
    """Get the edit scripts to transform between two given pronunciations.

    :param pron_a: Source pronunciation as list of strings, each string corresponding to a phoneme
    :param pron_b: Target pronunciation as list of strings, each string corresponding to a phoneme
    :param edit_costs: Costs of insert, replace and delete respectively
    :return: List of edit scripts.  Each edit script is represented as a list of operations,
                where each operation is a dictionary.
    """
    op_costs = {'insert': lambda x: edit_costs[0],
                'match': lambda x, y: 0 if x == y else edit_costs[1],
                'delete': lambda x: edit_costs[2]}

    distance, scripts, costs, ops = edit_distance.best_transforms(pron_a, pron_b, op_costs=op_costs)

    return [script.to_primitive() for script in scripts]


def print_edit_script(edit_script):
    """Print an edit script to the terminal.

    :param edit_script: The edit script as a list of operations, where each operation is a dictionary.
    """
    print('{}\n{}'.format(*edit_script_to_strings(edit_script)))


def edit_script_to_strings(edit_script, use_colors=True):
    """Convert an edit script to a pair of strings representing the operation in a human readable way.

    :param edit_script: The edit script as a list of operations, where each operation is a dictionary.
    :param use_colors: Boolean indicating whether to use terminal color codes to color the output.
    :return: Tuple with text corresponding to the first pronunciation and the text of the second one.
    """
    colors = collections.defaultdict(str)

    if use_colors:
        colors['red'] = '\x1b[31m'
        colors['normal'] = '\x1b[m'
        colors['green'] = '\x1b[32m'
        colors['on_red'] = '\x1b[41m'

    src_txt = ''
    dst_txt = ''
    for op in edit_script:
        if op['op_code'] == 'match':
            width = max(len(op['from_symbol']), len(op['to_symbol']))
            if op['from_symbol'] == op['to_symbol']:
                src_txt += '{green}{from_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                            op,
                                                                                            {'width': width}))
                dst_txt += '{green}{to_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                          op,
                                                                                          {'width': width}))
            else:
                src_txt += '{red}{from_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                          op,
                                                                                          {'width': width}))
                dst_txt += '{red}{to_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                        op,
                                                                                        {'width': width}))

        elif op['op_code'] == 'insert':
            space = ' '*len(op['to_symbol'])
            src_txt += '{on_red}{space}{normal}'.format(space=space, **_combine_dicts(colors,  op))
            dst_txt += '{red}{to_symbol}{normal}'.format(**_combine_dicts(colors, op))

        elif op['op_code'] == 'delete':
            space = ' '*len(op['from_symbol'])
            src_txt += '{red}{from_symbol}{normal}'.format(**_combine_dicts(colors, op))
            dst_txt += '{on_red}{space}{normal}'.format(space=space, **_combine_dicts(colors, op))

        src_txt += ' '
        dst_txt += ' '

    return src_txt, dst_txt




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import sys
import json

import numpy


class Operation:
    def __init__(self, op_code, index, from_symbol, to_symbol):
        self.op_code = op_code
        self.index = index
        self.from_symbol = from_symbol
        self.to_symbol = to_symbol

    def to_primitive(self):
        return self.__dict__

    def __repr__(self):
        return u'({}, {}, {}, {})'.format(self.op_code, self.index, self.from_symbol, self.to_symbol)


class Script:
    def __init__(self, i, j, path=None):
        self.operations = copy.copy(path.operations) if path is not None else []

        self.i = i
        self.j = j

        # The path cannot go any further
        if self.i < 0:
            pass

    def to_primitive(self):
        return [op.to_primitive() for op in reversed(self.operations)]

    def to_json(self):
        return json.dumps(self.to_primitive())

    def finished(self):
        return self.i < 0 and self.j < 0

    def print_colors(self):
        from blessings import Terminal

        term = Terminal()

        src_txt = u''
        dst_txt = u''
        for op in reversed(self.operations):
            if op.op_code == 'match':
                if op.from_symbol == op.to_symbol:
                    src_txt += u'{t.green}{op.from_symbol}{t.normal}'.format(t=term, op=op)
                    dst_txt += u'{t.green}{op.to_symbol}{t.normal}'.format(t=term, op=op)
                else:
                    src_txt += u'{t.red}{op.from_symbol}{t.normal}'.format(t=term, op=op)
                    dst_txt += u'{t.red}{op.to_symbol}{t.normal}'.format(t=term, op=op)

            elif op.op_code == 'insert':
                src_txt += u'{t.on_red} {t.normal}'.format(t=term, op=op)
                dst_txt += u'{t.red}{op.to_symbol}{t.normal}'.format(t=term, op=op)

            elif op.op_code == 'delete':
                src_txt += u'{t.red}{op.from_symbol}{t.normal}'.format(t=term, op=op)
                dst_txt += u'{t.on_red} {t.normal}'.format(t=term, op=op)

            src_txt += u' '
            dst_txt += u' '

        print(u'---\n{}\n{}\n---'.format(src_txt, dst_txt))

    def __repr__(self):
        return u'{}'.format([u'{}'.format(op) for op in self.operations[::-1] if not op.from_symbol == op.to_symbol])


def best_transforms(src, trg, op_costs=None):
    default_costs = {'match': lambda x, y: 0 if x == y else 1,
                     'insert': lambda x: 1,
                     'delete': lambda x: 1}

    op_costs = op_costs or default_costs

    INSERT = op_costs['insert']
    MATCH = op_costs['match']
    DELETE = op_costs['delete']

    src_len = len(src)
    trg_len = len(trg)

    # Initialize costs
    costs = numpy.ones((src_len, trg_len)) * numpy.inf

    op_type = numpy.dtype({'names': ['match',
                                     'insert',
                                     'delete'],
                           'formats': ['b',
                                       'b',
                                       'b']})
    ops = numpy.zeros((src_len, trg_len), dtype=op_type)

    s = numpy.array(src)
    t = numpy.array(trg)

    # Compute cost of 1st source item and 1st target item
    costs[0, 0] = MATCH(s[0], t[0])
    ops[0, 0]['match'] = True
    i, j = 0, 0

    # Compute costs of 1st source item and nth target item
    for j in xrange(1, costs.shape[1]):
        costs[0, j] = sum([INSERT(t[k]) for k in xrange(j)]) + MATCH(s[0], t[j])
        ops[0, j]['match'] = True

    # Compute costs of 1st target item and nth source item
    for i in xrange(1, costs.shape[0]):
        costs[i, 0] = sum([DELETE(s[k]) for k in xrange(i)]) + MATCH(s[i], t[0])
        ops[i, 0]['match'] = True

    # Compute costs of everything else
    for i in xrange(1, costs.shape[0]):
        for j in xrange(1, costs.shape[1]):
            # Replace cost
            cost = costs[i - 1, j - 1] + MATCH(s[i], t[j])
            if cost <= costs[i, j]:
                if cost < costs[i, j]:
                    ops[i, j] = (False, False, False)

                costs[i, j] = cost
                ops[i, j]['match'] = True

            # Delete cost
            cost = costs[i - 1, j] + DELETE(s[i])
            if cost <= costs[i, j]:
                if cost < costs[i, j]:
                    ops[i, j] = (False, False, False)

                costs[i, j] = cost
                ops[i, j]['delete'] = True

            # Insert cost
            cost = costs[i, j - 1] + INSERT(t[j])
            if cost <= costs[i, j]:
                if cost < costs[i, j]:
                    ops[i, j] = (False, False, False)

                costs[i, j] = cost
                ops[i, j]['insert'] = True

    # Compute transforms by backtracking paths
    paths = [Script(src_len - 1, trg_len - 1)]
    while not all(map(lambda x: x.finished(), paths)):
        new_paths = []

        for path in paths:
            # Check if we have aligned up to the begining of target or source
            if path.i < 0 or path.j < 0:
                # If we are in the top left corner of the cost matrix we are finished
                if path.finished():
                    new_paths.append(path)
                    continue

                # Else we have to add the insertions or deletions required to reach it
                else:
                    for k in xrange(path.i, -1, -1):
                        op = Operation('delete', k * 2 + 1, src[k], None)
                        path.operations.append(op)
                        path.i = k - 1

                    for k in xrange(path.j, -1, -1):
                        op = Operation('insert', 0, None, trg[k])
                        path.operations.append(op)
                        path.j = k - 1

                    new_paths.append(path)
                    continue

            if ops[path.i, path.j]['match']:
                op = Operation('match', path.i * 2 + 1, src[path.i], trg[path.j])
                new_path = Script(path.i - 1, path.j - 1, path=path)
                new_path.operations.append(op)
                new_paths.append(new_path)

            if ops[path.i, path.j]['delete']:
                op = Operation('delete', path.i * 2 + 1, src[path.i], None)
                new_path = Script(path.i - 1, path.j, path=path)
                new_path.operations.append(op)
                new_paths.append(new_path)

            if ops[path.i, path.j]['insert']:
                op = Operation('insert', path.i * 2 + 2, None, trg[path.j])
                new_path = Script(path.i, path.j - 1, path=path)
                new_path.operations.append(op)
                new_paths.append(new_path)

        paths = new_paths

    return costs[i, j], paths, costs, ops


if __name__ == '__main__':
    a = 'm u ch a'
    b = 'm u ch o s'

    a = u'k w i ð̞ a'
    b = u'i ð̞ a'

    a = u'x e s t o'
    b = u'p ɾ i n θ e s a'

    if len(sys.argv) > 2:
        a = sys.argv[1]
        b = sys.argv[2]

    src = a.split()
    trg = b.split()

    distance, transfs, costs, ops = best_transforms(src, trg)

    print('distance: {}'.format(distance))
    print('costs:\n{}'.format(costs))
    print('ops:\n{}'.format(ops))
    print(transfs[0].to_json())
    [transf.print_colors() for transf in transfs]
    
